//#include <unistd.h>
//#include <signal.h>
#include <iostream.h>

#include "signal_handler.hh"
#include "list.hh"

/************************************************************************
**                            SIGNAL_HANDLER                           **
************************************************************************/

List<Signal_Handler> Signal_Handler::list_handlers[MAX_SIGNALS][MAX_PRI_LEVELS];

int Signal_Handler::gstatus = SIGNALS_HANDLER_HANDLING_OFF;

int Signal_Handler::some_handler[MAX_SIGNALS] = {0,0,0,0,0,0,0,0,
						 0,0,0,0,0,0,0,0,
						 0,0,0,0,0,0,0,0,
						 0,0,0,0,0,0,0,0};

Signal_Handler::Signal_Handler(int _wich_signal,int _priority)
: wich_signal(_wich_signal), priority(_priority), status(SIGNAL_HANDLER_STATUS_OFF)
{
  //cout << "Signal_Handler: constructor\n";

  if((wich_signal < 0)||(wich_signal >= MAX_SIGNALS))
    {
      cerr << "Signal_Handler(): Uncorrect signal number. Exiting\n";
      exit(-1);
    }
  if((priority < 0)||(priority >= MAX_PRI_LEVELS))
    {
      cerr << "Signal_Handler(): Uncorrect level number. Exiting\n";
      exit(-1);
    }

  list_handlers[wich_signal][priority].AddHead(this);
}


Signal_Handler::~Signal_Handler()
{
  for(int i=1;i<=list_handlers[wich_signal][priority].Length();i++)
    if(&list_handlers[wich_signal][priority][i] == this)
      {
	if(status == SIGNAL_HANDLER_STATUS_ON)
	  TurnOffCatching();

	list_handlers[wich_signal][priority].Remove(i);
	if(list_handlers[wich_signal][priority].Length() == 0)
	  {
	    //Signal_Manager::SetDefaultAction(wich_signal); // no good for all
	    gstatus = SIGNALS_HANDLER_HANDLING_OFF;
	  }
	break;
      }
}

void Signal_Handler::TurnOnCatching()
{
  if(gstatus == SIGNALS_HANDLER_HANDLING_OFF)  
    {
      gstatus = SIGNALS_HANDLER_HANDLING_ON;
      cout << "Turn on catching globally\n";
    }
  
  if(some_handler[wich_signal] == 0)
    {
#ifdef __osf__
      Signal_Action  action(Signal_Handler::Sig_Handler,0x00);
#endif
#ifdef __sun__
      Signal_Action  action(Signal_Handler::Sig_Handler,0x00);
#endif
#ifdef __linux__
      Signal_Action  action(Signal_Handler::Sig_Handler,SYS_SA_MULTI);
#endif            
      Signal_Manager::SetAction(wich_signal,action);
      cout << "Turn on general catching for sig:" << wich_signal << '\n';
    }
  some_handler[wich_signal] += 1;
  
  status = SIGNAL_HANDLER_STATUS_ON;
  cout << "Turn on catching for signal handler:" << (void*) this <<  "\tsig:" << wich_signal << '\n';
}

void Signal_Handler::TurnOffCatching()
{
  status = SIGNAL_HANDLER_STATUS_OFF;
  cout << "Turn off catching for signal handler:" << (void*) this << '\n';

  if(some_handler[wich_signal] > 1)         // <----------- ( ... > 0)
    some_handler[wich_signal]--;
  else
    ;   // <---------------------- Set the default action
}

void Signal_Handler::Handler()
{
  cerr << "SIGNAL HANDLER: Default Signal Handler invoked: no action !\n";
}


/*********************  First Step  ************************/

void Signal_Handler::Sig_Handler(int sig)
{
  //cout << "SIGNAL HANDLER: received signal " << sig << '\n';

  //cout << "Possibles 0 Handlers = " << list_handlers[sig][0].Length() << '\n';
  //cout << "Possibles 1 Handlers = " << list_handlers[sig][1].Length() << '\n';
  //cout << "Possibles 2 Handlers = " << list_handlers[sig][2].Length() << '\n';
  //cout << "Possibles 3 Handlers = " << list_handlers[sig][3].Length() << '\n';

  for(int k=0;k<MAX_PRI_LEVELS;k++)
    for(int i=1;i<=list_handlers[sig][k].Length();i++)
      {
	//cout << "handler " << k << '.' << i << '\n';
	if(list_handlers[sig][k][i].status == SIGNAL_HANDLER_STATUS_ON)
	  {
	    //cout << "invoking handler " << i << '\n';
	    list_handlers[sig][k][i].Handler();
	  }
      }
	
  //cout << "SIGNAL HANDLER: finish to process signal " << sig << '\n';
}


